#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------
import json
import logging
import os
import sys
from typing import Dict

import torch
from pytorch_lightning import LightningDataModule, seed_everything

from health_azure.logging import logging_section, print_message_with_rank_pid
from health_azure.utils import (
    ENV_GLOBAL_RANK,
    ENV_LOCAL_RANK,
    ENV_NODE_RANK,
    ENV_OMPI_COMM_WORLD_RANK,
    PARENT_RUN_CONTEXT,
    RUN_CONTEXT,
    get_metrics_for_hyperdrive_run,
    get_metrics_for_run,
    is_global_rank_zero,
    is_local_rank_zero,
    is_running_in_azure_ml,
)
from health_ml.model_trainer import create_lightning_trainer, write_experiment_summary_file
from health_ml.runner_base import RunnerBase
from health_ml.utils.checkpoint_utils import cleanup_checkpoints
from health_ml.utils.common_utils import (
    change_working_directory,
)
from health_ml.utils.regression_test_utils import REGRESSION_TEST_METRICS_FILENAME, compare_folders_and_run_outputs


class TrainingRunner(RunnerBase):
    def get_data_module(self) -> LightningDataModule:
        return self.container.get_data_module()

    def get_multiple_trainloader_mode(self) -> str:
        # Workaround for a bug in PL 1.5.5: We need to pass the cycle mode for the training data as a trainer argument
        # because training data that uses a CombinedLoader is not split correctly in DDP. This flag cannot be passed
        # through the get_trainer_arguments method of the container because cycle mode is not yet available.
        multiple_trainloader_mode = "max_size_cycle"
        try:
            from SSL.data.datamodules import CombinedDataModule  # type: ignore

            if isinstance(self.data_module, CombinedDataModule):
                self.data_module.prepare_data()
                multiple_trainloader_mode = self.data_module.train_loader_cycle_mode  # type: ignore
                assert multiple_trainloader_mode, "train_loader_cycle_mode should be available now"
        except ModuleNotFoundError:
            pass
        return multiple_trainloader_mode

    def init_training(self) -> None:
        """
        Execute some bookkeeping tasks only once if running distributed and initialize the runner's trainer object.
        """
        if is_global_rank_zero():
            logging.info(f"Model checkpoints are saved at {self.container.checkpoint_folder}")
            write_experiment_summary_file(self.container, outputs_folder=self.container.outputs_folder)
            self.container.before_training_on_global_rank_zero()

        if is_local_rank_zero():
            self.container.before_training_on_local_rank_zero()
        self.container.before_training_on_all_ranks()

        # Set random seeds just before training. Ensure that dataloader workers are also seeded correctly.
        seed_everything(self.container.get_effective_random_seed(), workers=True)

        # Get the container's datamodule
        self.data_module = self.get_data_module()

        if not self.container.run_inference_only:
            checkpoint_path_for_recovery = self.checkpoint_handler.get_recovery_or_checkpoint_path_train()
            if not checkpoint_path_for_recovery and self.container.resume_training:
                # If there is no recovery checkpoint (e.g job hasn't been resubmitted) and a source checkpoint is given,
                # use it to resume training.
                checkpoint_path_for_recovery = self.checkpoint_handler.trained_weights_path

            self.trainer, self.storing_logger = create_lightning_trainer(
                container=self.container,
                resume_from_checkpoint=checkpoint_path_for_recovery,
                num_nodes=self.container.num_nodes,
                multiple_trainloader_mode=self.get_multiple_trainloader_mode(),
                azureml_run_for_logging=self.azureml_run_for_logging,
            )

            rank_info = ", ".join(
                f"{env}: {os.getenv(env)}" for env in [ENV_GLOBAL_RANK, ENV_LOCAL_RANK, ENV_NODE_RANK]
            )
            logging.info(f"Environment variables: {rank_info}. trainer.global_rank: {self.trainer.global_rank}")

    def after_ddp_cleanup(self, environ_before_training: Dict) -> None:
        """
        Run processes cleanup after ddp context to prepare for single device inference.
        Kill all processes in DDP besides rank 0.
        """

        # DDP will start multiple instances of the runner, one for each GPU. Those should terminate here after training.
        # We can now use the global_rank of the Lightning model, rather than environment variables, because DDP has set
        # all necessary properties.
        if self.container.model.global_rank != 0:
            print_message_with_rank_pid(
                f"Terminating training thread with rank {self.container.model.global_rank}.", level='INFO'
            )
            sys.exit()

        # Lightning modifies a ton of environment variables. If we first run training and then the test suite,
        # those environment variables will mislead the training runs in the test suite, and make them crash.
        # Hence, restore the original environment after training.
        os.environ.clear()
        os.environ.update(environ_before_training)

        if ENV_OMPI_COMM_WORLD_RANK in os.environ:
            del os.environ[ENV_OMPI_COMM_WORLD_RANK]

        # From the training setup, torch still thinks that it should run in a distributed manner,
        # and would block on some GPU operations. Hence, clean up distributed training.
        if torch.distributed.is_initialized():  # type: ignore
            torch.distributed.destroy_process_group()  # type: ignore

    def end_training(self, environ_before_training: Dict) -> None:
        """Cleanup after training is done. This is called after the trainer has finished fitting the data.
        This is called to update the checkpoint handler state and remove redundant checkpoint files. If running
        inference on a single device, this is also called to kill all processes besides rank 0.
        """
        # Update the checkpoint handler state
        self.checkpoint_handler.additional_training_done()
        if self.container.model.global_rank == 0:
            logging.info("Removing redundant checkpoint files.")
            cleanup_checkpoints(self.container.checkpoint_folder)

        if self.container.max_num_gpus_inference == 1:
            # Kill all processes besides rank 0 after training is done to start inference on a single device
            self.after_ddp_cleanup(environ_before_training)

    def is_crossval_disabled_or_child_0(self) -> bool:
        """
        Returns True if the present run is a non-cross-validation run, or child run 0 of a cross-validation run.
        """
        if self.container.is_crossvalidation_child_run:
            return self.container.crossval_index == 0
        return True

    def run_training(self) -> None:
        """
        The main training loop. It creates the Pytorch model based on the configuration options passed in,
        creates a Pytorch Lightning trainer, and trains the model.
        If a checkpoint was specified, then it loads the checkpoint before resuming training.
        The cwd is changed to the outputs folder so that the model can write to current working directory, and still
        everything is put into the right place in AzureML (only the contents of the "outputs" folder is treated as a
        result file).
        """
        with change_working_directory(self.container.outputs_folder):
            assert self.trainer, "Trainer should be initialized before training. Call self.init_training() first."
            self.trainer.fit(self.container.model, datamodule=self.data_module)
        for logger in self.trainer.loggers:
            assert logger is not None
            logger.finalize('success')

    def init_inference(self) -> None:
        """
        Prepare the trainer for running inference on the validation and test set. This chooses a checkpoint,
        initializes the PL Trainer object, and chooses the right data module. The hook for running
        inference on the validation set is run (`LightningContainer.on_run_extra_validation_epoch`) is first called to
        reflect any changes to the model or datamodule states before running inference.
        """
        if self.container.run_extra_val_epoch:
            logging.info("Preparing to run an extra validation epoch to evaluate the model on the validation set.")
            self.container.on_run_extra_validation_epoch()
        super().init_inference()

    def run_validation(self) -> None:
        """Run validation on the validation set for all models to save time/memory consuming outputs. This is done in
        inference only mode or when the user has requested an extra validation epoch. The cwd is changed to the outputs
        folder"""
        if self.container.run_extra_val_epoch or self.container.run_inference_only:
            with change_working_directory(self.container.outputs_folder):
                assert self.trainer, "Trainer should be initialized before validation. Call self.init_inference()."
                self.trainer.validate(
                    self.container.model, datamodule=self.data_module, ckpt_path=self.inference_checkpoint
                )
        else:
            logging.info("Skipping extra validation because the user has not requested it.")

    def run_regression_test(self) -> None:
        if self.container.regression_test_folder:
            with logging_section("Regression Test"):
                # Comparison with stored results for cross-validation runs only operates on child run 0. This run
                # has usually already downloaded the results for the other runs, and uploaded files to the parent
                # run context.
                regression_metrics_str = self.container.regression_metrics
                regression_metrics = regression_metrics_str.split(',') if regression_metrics_str else None
                # TODO: user should be able to override this value
                crossval_arg_name = self.container.CROSSVAL_INDEX_ARG_NAME

                logging.info("Comparing the current results against stored results.")
                if self.is_crossval_disabled_or_child_0():
                    if is_running_in_azure_ml():
                        if PARENT_RUN_CONTEXT is not None:
                            metrics = get_metrics_for_hyperdrive_run(
                                child_run_arg_name=crossval_arg_name,
                                run=PARENT_RUN_CONTEXT,
                                keep_metrics=regression_metrics,
                            )
                        else:
                            metrics = get_metrics_for_run(run=RUN_CONTEXT, keep_metrics=regression_metrics)

                        if metrics:
                            metrics_filename = self.container.outputs_folder / REGRESSION_TEST_METRICS_FILENAME
                            logging.info(f"Saving metrics to {metrics_filename}")
                            metrics_filename.write_text(json.dumps(metrics))

                    compare_folders_and_run_outputs(
                        expected=self.container.regression_test_folder,
                        actual=self.container.outputs_folder,
                        csv_relative_tolerance=self.container.regression_test_csv_tolerance,
                    )
                else:
                    logging.info("Skipping as this is not cross-validation child run 0.")
        else:
            logging.info("Skipping regression test, no available results to compare with.")

    def run(self) -> None:
        """
        Driver function to run a ML experiment
        """
        self.init_training()

        if not self.container.run_inference_only:
            # Backup the environment variables in case we need to run a second training in the unit tests.
            environ_before_training = dict(os.environ)

            with logging_section("Model training"):
                self.run_training()

            self.end_training(environ_before_training)

        self.init_inference()

        with logging_section("Model validation"):
            self.run_validation()

        with logging_section("Model inference"):
            self.run_inference()

        self.run_regression_test()



#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------
""" Layers for attention-based multi-instance learning (DeepMIL)
Created using the original DeepMIL paper and code from Ilse et al., 2018
https://github.com/AMLab-Amsterdam/AttentionDeepMIL (MIT License)
"""
from typing import Tuple, Optional
from torch import nn, Tensor, transpose, mm
import torch
import torch.nn.functional as F
from torch.nn import Module, TransformerEncoderLayer


class MeanPoolingLayer(nn.Module):
    """Mean pooling returns uniform weights and the average feature vector over the first axis"""

    def forward(self, features: Tensor) -> Tuple[Tensor, Tensor]:
        num_instances = features.shape[0]
        attention_weights = torch.full((1, num_instances), 1.0 / num_instances)
        pooled_features = features.mean(dim=0)
        pooled_features = pooled_features.view(1, -1)
        return (attention_weights, pooled_features)


class MaxPoolingLayer(nn.Module):
    """Max pooling returns uniform weights and the maximum feature vector over the first axis"""

    def forward(self, features: Tensor) -> Tuple[Tensor, Tensor]:
        num_instances = features.shape[0]
        pooled_features, indices = features.max(dim=0)
        frequency = torch.bincount(indices, minlength=num_instances)
        frequency_norm = frequency / sum(frequency)
        attention_weights = frequency_norm.view(1, num_instances)
        pooled_features = pooled_features.view(1, -1)
        return (attention_weights, pooled_features)


class AttentionLayer(nn.Module):
    """AttentionLayer: Simple attention layer
    Requires size of input L, hidden D, and attention layers K (default K=1)
    """

    def __init__(self, input_dims: int, hidden_dims: int, attention_dims: int = 1) -> None:
        super().__init__()

        # Attention layers
        self.input_dims = input_dims  # L
        self.hidden_dims = hidden_dims  # D
        self.attention_dims = attention_dims  # K
        self.attention = nn.Sequential(
            nn.Linear(self.input_dims, self.hidden_dims), nn.Tanh(), nn.Linear(self.hidden_dims, self.attention_dims)
        )

    def forward(self, features: Tensor) -> Tuple[Tensor, Tensor]:
        features = features.view(-1, self.input_dims)  # N x L
        attention_weights = self.attention(features)  # N x K
        attention_weights = transpose(attention_weights, 1, 0)  # K x N
        attention_weights = F.softmax(attention_weights, dim=1)  # Softmax over N : K x N
        pooled_features = mm(attention_weights, features)  # Matrix multiplication : K x L
        return attention_weights, pooled_features


class GatedAttentionLayer(nn.Module):
    """GatedAttentionLayer: Gated attention layer
    Requires size of input L, hidden D, and output layers K (default K=1)
    """

    def __init__(self, input_dims: int, hidden_dims: int, attention_dims: int = 1) -> None:
        super().__init__()

        # Gated attention layers
        self.input_dims = input_dims  # L
        self.hidden_dims = hidden_dims  # D
        self.attention_dims = attention_dims  # K
        self.attention_V = nn.Sequential(nn.Linear(self.input_dims, self.hidden_dims), nn.Tanh())
        self.attention_U = nn.Sequential(nn.Linear(self.input_dims, self.hidden_dims), nn.Sigmoid())
        self.attention_weights = nn.Linear(self.hidden_dims, self.attention_dims)

    def forward(self, features: Tensor) -> Tuple[Tensor, Tensor]:
        features = features.view(-1, self.input_dims)  # N x L
        A_V = self.attention_V(features)  # N x D
        A_U = self.attention_U(features)  # N x D
        attention_weights = self.attention_weights(A_V * A_U)  # Element-wise multiplication : N x K
        attention_weights = transpose(attention_weights, 1, 0)  # K x N
        attention_weights = F.softmax(attention_weights, dim=1)  # Softmax over N : K x N
        pooled_features = mm(attention_weights, features)  # Matrix multiplication : K x L
        return attention_weights, pooled_features


class CustomTransformerEncoderLayer(TransformerEncoderLayer):
    """Adaptation of the pytorch TransformerEncoderLayer that always outputs the attention weights.

    TransformerEncoderLayer is made up of self-attn and feedforward network.
    This standard encoder layer is based on the paper "Attention Is All You Need".
    Ashish Vaswani, Noam Shazeer, Niki Parmar, Jakob Uszkoreit, Llion Jones, Aidan N Gomez,
    Lukasz Kaiser, and Illia Polosukhin. 2017. Attention is all you need. In Advances in
    Neural Information Processing Systems, pages 6000-6010. Users may modify or implement
    in a different way during application.

    Args:
        d_model: the number of expected features in the input (required).
        nhead: the number of heads in the multiheadattention models (required).
        dim_feedforward: the dimension of the feedforward network model (default=2048).
        dropout: the dropout value (default=0.1).
        activation: the activation function of the intermediate layer, can be a string
            ("relu" or "gelu") or a unary callable. Default: relu
        layer_norm_eps: the eps value in layer normalization components (default=1e-5).
        batch_first: If ``True``, then the input and output tensors are provided
            as (batch, seq, feature). Default: ``False``.
        norm_first: if ``True``, layer norm is done prior to attention and feedforward
            operations, respectivaly. Otherwise it's done after. Default: ``False`` (after).

    Examples::
        >>> encoder_layer = nn.TransformerEncoderLayer(d_model=512, nhead=8)
        >>> src = torch.rand(10, 32, 512)
        >>> out, attention_weights = encoder_layer(src)

    Alternatively, when ``batch_first`` is ``True``:
        >>> encoder_layer = nn.TransformerEncoderLayer(d_model=512, nhead=8, batch_first=True)
        >>> src = torch.rand(32, 10, 512)
        >>> out, attention_weights = encoder_layer(src)
    """

    # new forward returns output as well as attention weights

    def forward(
        self,
        src: torch.Tensor,  # type: ignore
        src_mask: Optional[torch.Tensor] = None,
        src_key_padding_mask: Optional[torch.Tensor] = None,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Pass the input through the encoder layer.

        Args:
            src: the sequence to the encoder layer (required).
            src_mask: the mask for the src sequence (optional).
            src_key_padding_mask: the mask for the src keys per batch (optional).

        Shape:
            see the docs in Transformer class.
        """

        # see Fig. 1 of https://arxiv.org/pdf/2002.04745v1.pdf

        x = src
        if self.norm_first:
            sa_block_out, a = self._sa_block(self.norm1(x), src_mask, src_key_padding_mask)
            x = x + sa_block_out
            x = x + self._ff_block(self.norm2(x))
        else:
            sa_block_out, a = self._sa_block(x, src_mask, src_key_padding_mask)
            x = self.norm1(x + sa_block_out)
            x = self.norm2(x + self._ff_block(x))

        return x, a

    # new self-attention block, returns output as well as attention weights
    def _sa_block(
        self, x: Tensor, attn_mask: Optional[Tensor], key_padding_mask: Optional[Tensor]  # type: ignore
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        x, a = self.self_attn(
            x, x, x, attn_mask=attn_mask, key_padding_mask=key_padding_mask, need_weights=True
        )  # Just because of this flag I had to copy all of the code...
        x = x[0]
        return self.dropout1(x), a


class TransformerPooling(Module):
    """Create a Transformer encoder module consisting of multiple Transformer encoder layers.

    We use a additional classification token (cls token) for pooling like seen in ViT/Bert. First, the cls token is
    appended to the list of tiles encodings. Second, we perform self-attention between all tile encodings and the cls
    token. Last, we extract the cls token and use it for classification.

    Args:
        num_layers: Number of Transformer encoder layers.
        num_heads: Number of attention heads per layer.
        dim_representation: Dimension of input encoding.
        transformer_dropout: The dropout value of transformer encoder layers.
    """

    def __init__(self, num_layers: int, num_heads: int, dim_representation: int, transformer_dropout: float) -> None:
        super(TransformerPooling, self).__init__()
        self.num_layers = num_layers
        self.num_heads = num_heads
        self.dim_representation = dim_representation
        self.transformer_dropout = transformer_dropout
        self.cls_token = nn.Parameter(torch.zeros([1, dim_representation]))

        self.transformer_encoder_layers = []
        for _ in range(self.num_layers):
            self.transformer_encoder_layers.append(
                CustomTransformerEncoderLayer(
                    self.dim_representation,
                    self.num_heads,
                    dim_feedforward=self.dim_representation,
                    dropout=self.transformer_dropout,
                    activation=F.gelu,
                    batch_first=True,
                )
            )
        self.transformer_encoder_layers = torch.nn.ModuleList(self.transformer_encoder_layers)  # type: ignore

    def forward(self, features: Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        # Append cls token
        features = torch.vstack([self.cls_token, features]).unsqueeze(0)

        for i in range(self.num_layers):
            features, attention_weights = self.transformer_encoder_layers[i](features)

        # Extract cls token
        pooled_features = features[:, 0]

        # Get attention weights with respect to the cls token, without the element where it attends to itself

        self_attention_cls_token = attention_weights[0, 0, 0]  # type: ignore
        attention_weights = attention_weights[:, 0, 1:]  # type: ignore

        # We want A to sum to one, simple hack: add self_attention_cls_token/num_tiles to each element
        attention_weights += self_attention_cls_token / attention_weights.shape[-1]

        return (attention_weights, pooled_features)


class TransformerPoolingBenchmark(Module):
    """Create a Transformer encoder module consisting of multiple Transformer encoder layers.
    The pooling is inspired by the transformer pooling in `monai.networks.nets.milmodel`.
    The transformer pooling is used in the implementation of (Myronenko et al. 2021).
    Example in https://github.com/Project-MONAI/tutorials/blob/master/pathology/multiple_instance_learning/
    panda_mil_train_evaluate_pytorch_gpu.py.
    Args:
        num_layers: Number of Transformer encoder layers.
        num_heads: Number of attention heads per layer.
        dim_representation: Dimension of input encoding.
        transformer_dropout: The dropout value of transformer encoder layers.
    """

    def __init__(
        self, num_layers: int, num_heads: int, dim_representation: int, hidden_dim: int, transformer_dropout: float
    ) -> None:
        super().__init__()
        self.num_layers = num_layers
        self.num_heads = num_heads
        self.dim_representation = dim_representation
        self.hidden_dim = hidden_dim
        self.transformer_dropout = transformer_dropout
        transformer_layer = nn.TransformerEncoderLayer(
            d_model=self.dim_representation, nhead=self.num_heads, dropout=self.transformer_dropout, batch_first=True
        )
        self.transformer = nn.TransformerEncoder(transformer_layer, num_layers=self.num_layers)
        self.attention = nn.Sequential(
            nn.Linear(self.dim_representation, self.hidden_dim), nn.Tanh(), nn.Linear(self.hidden_dim, 1)
        )

    def forward(self, x: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Input size is L, bag size N, hidden dimension is D, and attention layers K (default K=1).
        """
        x = x.reshape(-1, x.shape[0], x.shape[1])  # 1 x N X L
        x = self.transformer(x)  # 1 x N X L
        a = self.attention(x)  # 1 x N X K
        attention_weights = torch.softmax(a, dim=1)  # 1 x N x K
        pooled_features = torch.sum(x * attention_weights, dim=1)  # K X L
        attention_weights = attention_weights.permute(0, 2, 1)  # 1 x K X N
        return (attention_weights.squeeze(0), pooled_features)  # K X N, K X L



#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------
from __future__ import annotations

import random
from dataclasses import dataclass
from itertools import combinations
from math import ceil
from typing import Any, Dict, Iterable, Optional, Sequence, Set, Tuple, List

import numpy as np
import pandas as pd
from sklearn.model_selection import GroupKFold, KFold

from health_ml.utils import common_utils
from health_ml.utils.common_utils import ModelExecutionMode


@dataclass
class DatasetSplits:
    train: pd.DataFrame
    val: pd.DataFrame
    test: pd.DataFrame
    subject_column: Optional[str] = None
    group_column: Optional[str] = None
    allow_empty: bool = False

    def __post_init__(self) -> None:
        common_utils.check_properties_are_not_none(self)

        def pairwise_intersection(*collections: Iterable) -> Set:
            """
            Returns any element that appears in more than one collection

            :return: a Set of elements that appear in more than one collection
            """
            intersection = set()
            for col1, col2 in combinations(map(set, collections), 2):  # type: ignore
                intersection |= col1 & col2
            return intersection

        # perform dataset split validity assertions
        unique_train, unique_test, unique_val = self.unique_subjects()
        intersection = pairwise_intersection(unique_train, unique_test, unique_val)

        if len(intersection) != 0:
            raise ValueError("Train, Test, and Val splits must have no intersection, found: {}".format(intersection))

        if self.group_column is not None:
            groups_train = self.train[self.group_column].unique()
            groups_test = self.test[self.group_column].unique()
            groups_val = self.val[self.group_column].unique()
            group_intersection = pairwise_intersection(groups_train, groups_test, groups_val)
            if len(group_intersection) != 0:
                raise ValueError(
                    "Train, Test, and Val splits must have no intersecting groups, found: {}".format(group_intersection)
                )

        if (not self.allow_empty) and any([len(x) == 0 for x in [unique_train, unique_val]]):
            raise ValueError(
                "train_ids({}), val_ids({}) must have at least one value".format(len(unique_train), len(unique_val))
            )

    def __str__(self) -> str:
        unique_train, unique_test, unique_val = self.unique_subjects()
        return (
            f'Train: {len(unique_train)}, Test: {len(unique_test)}, and Val: {len(unique_val)}. '
            f'Total subjects: {len(unique_train) + len(unique_test) + len(unique_val)}'
        )

    def unique_subjects(self) -> Tuple[Any, Any, Any]:
        """
        Return a tuple of pandas Series of unique subjects across train, test and validation data splits,
        based on self.subject_column

        :return: a tuple of pandas Series
        """
        return (
            self.train[self.subject_column].unique(),
            self.test[self.subject_column].unique(),
            self.val[self.subject_column].unique(),
        )

    def number_of_subjects(self) -> int:
        """
        Returns the sum of unique subjects in the dataset (identified by self.subject_column), summed
        over train, test and validation data splits

        :return: An integer representing the number of unique subjects
        """
        unique_train, unique_test, unique_val = self.unique_subjects()
        return len(unique_train) + len(unique_test) + len(unique_val)

    def __getitem__(self, mode: ModelExecutionMode) -> pd.DataFrame:
        """
        Retrieve either the train, validation or test data in the form of a Pandas dataframe, depending
        on the current execution mode

        :param mode: The current ModelExecutionMode
        :return: A dataframe of the relevant data split
        """
        if mode == ModelExecutionMode.TRAIN:
            return self.train
        elif mode == ModelExecutionMode.TEST:
            return self.test
        elif mode == ModelExecutionMode.VAL:
            return self.val
        else:
            raise ValueError(f"Model execution mode not recognized: {mode}")

    @staticmethod
    def get_subject_ranges_for_splits(
        population: Sequence[str], proportion_train: float, proportion_test: float, proportion_val: float
    ) -> Dict[ModelExecutionMode, Set[str]]:
        """
        Get mutually exclusive subject ranges for each dataset split (w.r.t to the proportion provided)
        ensuring all sets have at least one item in them when possible.

        :param population: all subjects
        :param proportion_train: proportion for the train set.
        :param proportion_test: proportion for the test set.
        :param proportion_val: proportion for the validation set.
        :return: Train, Test, and Val splits
        """
        sum_proportions = proportion_train + proportion_val + proportion_test
        if not np.isclose(sum_proportions, 1):
            raise ValueError(
                "proportion_train({}) + proportion_val({}) + proportion_test({}) must be ~ 1, found: {}".format(
                    proportion_train, proportion_val, proportion_test, sum_proportions
                )
            )

        if not 0 <= proportion_test < 1:
            raise ValueError("proportion_test({}) must be in range [0, 1)".format(proportion_test))

        if not all([0 < x < 1 for x in [proportion_train, proportion_val]]):
            raise ValueError(
                "proportion_train({}) and proportion_val({}) must be in range (0, 1)".format(
                    proportion_train, proportion_val
                )
            )

        subjects_train, subjects_test, subjects_val = (set(population[0:1]), set(population[1:2]), set(population[2:3]))
        remaining = list(population[3:])
        if proportion_test == 0:
            remaining = list(subjects_test) + remaining
            subjects_test = set()

        subjects_train |= set(remaining[: ceil(len(remaining) * proportion_train)])
        if len(subjects_test) > 0:
            subjects_test |= set(
                remaining[len(subjects_train) : len(subjects_train) + ceil(len(remaining) * proportion_test)]
            )
        subjects_val |= set(remaining) - (subjects_train | subjects_test)
        result = {
            ModelExecutionMode.TRAIN: subjects_train,
            ModelExecutionMode.TEST: subjects_test,
            ModelExecutionMode.VAL: subjects_val,
        }
        return result

    @staticmethod
    def _from_split_keys(
        df: pd.DataFrame,
        train_keys: Sequence[str],
        test_keys: Sequence[str],
        val_keys: Sequence[str],
        *,  # make column names keyword-only arguments to avoid mistakes when providing both
        key_column: str,
        subject_column: str,
        group_column: Optional[str],
    ) -> DatasetSplits:
        """
        Takes a slice of values from each data split train/test/val for the provided keys.

        :param df: the input DataFrame
        :param train_keys: keys for training.
        :param test_keys: keys for testing.
        :param val_keys: keys for validation.
        :param key_column: name of the column the provided keys belong to
        :param subject_column: subject id column name
        :param group_column: grouping column name; if given, samples from each group will always be
            in the same subset (train, val, or test) and cross-validation fold.
        :return: Data splits with respected dataset split ids.
        """
        train_df = DatasetSplits.get_df_from_ids(df, train_keys, key_column)
        test_df = DatasetSplits.get_df_from_ids(df, test_keys, key_column)
        val_df = DatasetSplits.get_df_from_ids(df, val_keys, key_column)

        return DatasetSplits(
            train=train_df, test=test_df, val=val_df, subject_column=subject_column, group_column=group_column
        )

    @staticmethod
    def from_proportions(
        df: pd.DataFrame,
        proportion_train: float,
        proportion_test: float,
        proportion_val: float,
        *,  # make column names keyword-only arguments to avoid mistakes when providing both
        subject_column: str = "",
        group_column: Optional[str] = None,
        shuffle: bool = True,
        random_seed: int = 0,
    ) -> DatasetSplits:
        """
        Creates a split of a dataset into train, test, and validation set, according to fixed proportions using
        the "subject" column in the dataframe, or the group column, if given.

        :param df: The dataframe containing all subjects.
        :param proportion_train: proportion for the train set.
        :param proportion_test: proportion for the test set.
        :param subject_column: Subject id column name
        :param group_column: grouping column name; if given, samples from each group will always be
            in the same subset (train, val, or test) and cross-validation fold.
        :param proportion_val: proportion for the validation set.
        :param shuffle: If True the subjects in the dataframe will be shuffle before performing splits.
        :param random_seed: Random seed to be used for shuffle 0 is default.
        :return:
        """
        key_column: str = subject_column if group_column is None else group_column
        split_keys = df[key_column].unique()
        if shuffle:
            # fix the random seed so we can guarantee reproducibility when working with shuffle
            random.Random(random_seed).shuffle(split_keys)
        ranges = DatasetSplits.get_subject_ranges_for_splits(
            split_keys,
            proportion_train=proportion_train,
            proportion_val=proportion_val,
            proportion_test=proportion_test,
        )
        return DatasetSplits._from_split_keys(
            df,
            list(ranges[ModelExecutionMode.TRAIN]),
            list(ranges[ModelExecutionMode.TEST]),
            list(ranges[ModelExecutionMode.VAL]),
            key_column=key_column,
            subject_column=subject_column,
            group_column=group_column,
        )

    @staticmethod
    def from_subject_ids(
        df: pd.DataFrame,
        train_ids: Sequence[str],
        test_ids: Sequence[str],
        val_ids: Sequence[str],
        *,  # make column names keyword-only arguments to avoid mistakes when providing both
        subject_column: str = "",
        group_column: Optional[str] = None,
    ) -> DatasetSplits:
        """
        Assuming a DataFrame with columns subject
        Takes a slice of values from each data split train/test/val for the provided ids.

        :param df: the input DataFrame
        :param train_ids: ids for training.
        :param test_ids: ids for testing.
        :param val_ids: ids for validation.
        :param subject_column: subject id column name
        :param group_column: grouping column name; if given, samples from each group will always be
            in the same subset (train, val, or test) and cross-validation fold.
        :return: Data splits with respected dataset split ids.
        """
        return DatasetSplits._from_split_keys(
            df,
            train_ids,
            test_ids,
            val_ids,
            key_column=subject_column,
            subject_column=subject_column,
            group_column=group_column,
        )

    @staticmethod
    def from_groups(
        df: pd.DataFrame,
        train_groups: Sequence[str],
        test_groups: Sequence[str],
        val_groups: Sequence[str],
        *,  # make column names keyword-only arguments to avoid mistakes when providing both
        group_column: str,
        subject_column: str = "",
    ) -> DatasetSplits:
        """
        Assuming a DataFrame with columns subject
        Takes a slice of values from each data split train/test/val for the provided groups.

        :param df: the input DataFrame
        :param train_groups: groups for training.
        :param test_groups: groups for testing.
        :param val_groups: groups for validation.
        :param subject_column: subject id column name
        :param group_column: grouping column name; if given, samples from each group will always be
            in the same subset (train, val, or test) and cross-validation fold.
        :return: Data splits with respected dataset split ids.
        """
        return DatasetSplits._from_split_keys(
            df,
            train_groups,
            test_groups,
            val_groups,
            key_column=group_column,
            subject_column=subject_column,
            group_column=group_column,
        )

    @staticmethod
    def get_df_from_ids(df: pd.DataFrame, ids: Sequence[str], subject_column: Optional[str] = "") -> pd.DataFrame:
        """
        Retrieve a subset dataframe where the subject column is restricted to a sequence of provided ids

        :param df: The dataframe to restrict
        :param ids: The ids to lookup
        :param subject_column: The column to lookup ids in. Defaults to ""
        :return: A subset of the dataframe
        """
        return df[df[subject_column].isin(ids)]

    def get_k_fold_cross_validation_splits(self, n_splits: int, random_seed: int = 0) -> List[DatasetSplits]:
        """
        Creates K folds from the Train + Val splits.
        If a group_column has been specified, the folds will be split such that
        subjects in a group will not be separated. In this case, the splits are
        fully deterministic, and random_seed is ignored.

        :param n_splits: number of folds to perform.
        :param random_seed: random seed to be used for shuffle 0 is default.
        :return: List of K dataset splits
        """
        if n_splits <= 0:
            raise ValueError("n_splits must be >= 0 found {}".format(n_splits))
        # concatenate train and val, as training set = train + val
        cv_dataset = pd.concat([self.train, self.val])
        if self.group_column is None:  # perform standard subject-based k-fold cross-validation
            # unique subjects
            subject_ids = cv_dataset[self.subject_column].unique()
            # calculate the random split indices
            k_folds = KFold(n_splits=n_splits, shuffle=True, random_state=random_seed)
            folds_gen = k_folds.split(subject_ids)
        else:  # perform grouped k-fold cross-validation
            # Here we take all entries, rather than unique, to keep subjects
            # matched to groups in the resulting arrays. This works, but could
            # perhaps be improved with group-by logic...?
            subject_ids = cv_dataset[self.subject_column].values
            groups = cv_dataset[self.group_column].values
            # scikit-learn uses a deterministic algorithm for grouped splits
            # that tries to balance the group sizes in all folds
            k_folds = GroupKFold(n_splits=n_splits)
            folds_gen = k_folds.split(subject_ids, groups=groups)

        def ids_from_indices(indices: Sequence[int]) -> List[str]:
            return [subject_ids[x] for x in indices]

        # create the number of requested splits of the dataset
        return [
            DatasetSplits(
                train=self.get_df_from_ids(cv_dataset, ids_from_indices(train_indices), self.subject_column),
                val=self.get_df_from_ids(cv_dataset, ids_from_indices(val_indices), self.subject_column),
                test=self.test,
                subject_column=self.subject_column,
                group_column=self.group_column,
            )
            for train_indices, val_indices in folds_gen
        ]
